import time
from logging import getLogger

from aspectlib import Aspect

logger = getLogger(__name__)


def retry(func=None, retries=5, backoff=None, exceptions=(IOError, OSError, EOFError), cleanup=None, sleep=time.sleep):
    """
    Decorator that retries the call ``retries`` times if ``func`` raises ``exceptions``. Can use a ``backoff`` function
    to sleep till next retry.

    Example::

        >>> should_fail = lambda foo=[1,2,3]: foo and foo.pop()
        >>> @retry
        ... def flaky_func():
        ...     if should_fail():
        ...         raise OSError('Tough luck!')
        ...     print("Success!")
        ...
        >>> flaky_func()
        Success!

    If it reaches the retry limit::

        >>> @retry
        ... def bad_func():
        ...     raise OSError('Tough luck!')
        ...
        >>> bad_func()
        Traceback (most recent call last):
        ...
        OSError: Tough luck!

    """

    @Aspect(bind=True)
    def retry_aspect(cutpoint, *args, **kwargs):
        for count in range(retries + 1):
            try:
                if count and cleanup:
                    cleanup(*args, **kwargs)
                yield
                break
            except exceptions as exc:
                if count == retries:
                    raise
                if not backoff:
                    timeout = 0
                elif isinstance(backoff, (int, float)):
                    timeout = backoff
                else:
                    timeout = backoff(count)
                logger.exception("%s(%s, %s) raised exception %s. %s retries left. Sleeping %s secs.",
                                 cutpoint.__name__, args, kwargs, exc, retries - count, timeout)
                sleep(timeout)

    return retry_aspect if func is None else retry_aspect(func)


def exponential_backoff(count):
    """
    Wait 2**N seconds.
    """
    return 2 ** count


retry.exponential_backoff = exponential_backoff


def straight_backoff(count):
    """
    Wait 1, 2, 5 seconds. All retries after the 3rd retry will wait 5*N-5 seconds.
    """
    return (1, 2, 5)[count] if count < 3 else 5 * count - 5


retry.straight_backoff = straight_backoff


def flat_backoff(count):
    """
    Wait 1, 2, 5, 10, 15, 30 and 60 seconds. All retries after the 5th retry will wait 60 seconds.
    """
    return (1, 2, 5, 10, 15, 30, 60)[count if count < 6 else -1]


retry.flat_backoff = flat_backoff



import asyncio
import os
import re
import socket
import warnings
from datetime import timedelta

import pytest
from process_tests import dump_on_error
from process_tests import wait_for_strings
from tornado import gen
from tornado import ioloop

import aspectlib
from aspectlib import debug
from aspectlib.test import mock
from aspectlib.test import record
from aspectlib.utils import PYPY

try:
    from StringIO import StringIO
except ImportError:
    from io import StringIO

LOG_TEST_SOCKET = r"""^\{_?socket(object)?\}.connect\(\('127.0.0.1', 1\)\) +<<< .*tests[\/]test_integrations.py:\d+:test_socket.*
\{_?socket(object)?\}.connect \~ raised .*(ConnectionRefusedError|error)\((10061|111), .*refused.*\)"""


def test_mock_builtin():
    with aspectlib.weave(open, mock('foobar')):
        assert open('???') == 'foobar'

    assert open(__file__) != 'foobar'


def test_mock_builtin_os():
    print(os.open.__name__)
    with aspectlib.weave('os.open', mock('foobar')):
        assert os.open('???') == 'foobar'

    assert os.open(__file__, 0) != 'foobar'


def test_record_warning():
    with aspectlib.weave('warnings.warn', record):
        warnings.warn('crap')
        assert warnings.warn.calls == [(None, ('crap',), {})]


@pytest.mark.skipif(not hasattr(os, 'fork'), reason="os.fork not available")
def test_fork():
    with aspectlib.weave('os.fork', mock('foobar')):
        pid = os.fork()
        if not pid:
            os._exit(0)
        assert pid == 'foobar'

    pid = os.fork()
    if not pid:
        os._exit(0)
    assert pid != 'foobar'


def test_socket(target=socket.socket):
    buf = StringIO()
    with aspectlib.weave(target, aspectlib.debug.log(print_to=buf, stacktrace=4, module=False), lazy=True):
        s = socket.socket()
        try:
            s.connect(('127.0.0.1', 1))
        except Exception:
            pass

    print(buf.getvalue())
    assert re.match(LOG_TEST_SOCKET, buf.getvalue())

    s = socket.socket()
    try:
        s.connect(('127.0.0.1', 1))
    except Exception:
        pass

    assert re.match(LOG_TEST_SOCKET, buf.getvalue())


def test_socket_as_string_target():
    test_socket(target='socket.socket')


def test_socket_meth(meth=socket.socket.close):
    calls = []
    with aspectlib.weave(meth, record(calls=calls)):
        s = socket.socket()
        assert s.close() is None
    assert calls == [(s, (), {})]
    del calls[:]

    s = socket.socket()
    assert s.close() is None
    assert calls == []


def test_socket_meth_as_string_target():
    test_socket_meth('socket.socket.close')


def test_socket_all_methods():
    buf = StringIO()
    with aspectlib.weave(socket.socket, aspectlib.debug.log(print_to=buf, stacktrace=False), lazy=True, methods=aspectlib.ALL_METHODS):
        socket.socket()

    assert "}.__init__ => None" in buf.getvalue()


@pytest.mark.skipif(not hasattr(os, 'fork') or PYPY, reason="os.fork not available or PYPY")
def test_realsocket_makefile():
    buf = StringIO()
    p = socket.socket()
    p.bind(('127.0.0.1', 0))
    p.listen(1)
    p.settimeout(1)
    pid = os.fork()

    if pid:
        with aspectlib.weave(
            ['socket.SocketIO', 'socket.socket'],
            aspectlib.debug.log(print_to=buf, stacktrace=False),
            lazy=True,
            methods=aspectlib.ALL_METHODS,
        ):
            s = socket.socket()
            s.settimeout(1)
            s.connect(p.getsockname())
            fh = s.makefile('rwb', buffering=0)
            fh.write(b"STUFF\n")
            fh.readline()

        with dump_on_error(buf.getvalue):
            wait_for_strings(
                buf.getvalue,
                0,
                "}.connect",
                "}.makefile",
                "}.write(",
                "}.send",
                "}.write =>",
                "}.readline()",
                "}.recv",
                "}.readline => ",
            )
    else:
        try:
            c, _ = p.accept()
            c.settimeout(1)
            f = c.makefile('rw', buffering=1)
            while f.readline():
                f.write('-\n')
        finally:
            os._exit(0)


def test_weave_os_module():
    calls = []

    with aspectlib.weave('os', record(calls=calls, extended=True), methods="getenv|walk"):
        os.getenv('BUBU', 'bubu')
        os.walk('.')

    assert calls == [(None, 'os.getenv', ('BUBU', 'bubu'), {}), (None, 'os.walk', ('.',), {})]


def test_decorate_asyncio_coroutine():
    buf = StringIO()

    @debug.log(print_to=buf, module=False, stacktrace=2, result_repr=repr)
    async def coro():
        await asyncio.sleep(0.01)
        return "result"

    loop = asyncio.new_event_loop()
    loop.run_until_complete(coro())
    output = buf.getvalue()
    print(output)
    assert 'coro => %r' % 'result' in output


def test_decorate_tornado_coroutine():
    buf = StringIO()

    @gen.coroutine
    @debug.log(print_to=buf, module=False, stacktrace=2, result_repr=repr)
    def coro():
        if hasattr(gen, 'Task'):
            yield gen.Task(loop.add_timeout, timedelta(microseconds=10))
        else:
            yield gen.sleep(0.01)
        return "result"

    asyncio_loop = asyncio.new_event_loop()
    try:
        get_event_loop = asyncio.get_event_loop
        asyncio.get_event_loop = lambda: asyncio_loop
        loop = ioloop.IOLoop.current()
        loop.run_sync(coro)
    finally:
        asyncio.get_event_loop = get_event_loop
    output = buf.getvalue()
    assert 'coro => %r' % 'result' in output



from aspectlib import test


class Foo(object):
    def bar(self):
        return 1


def test_fixture_1(weave):
    weave(Foo.bar, test.mock(2))
    assert Foo().bar() == 2


def test_fixture_2(weave):
    assert Foo().bar() == 1

    with weave(Foo.bar, test.mock(2)):
        assert Foo().bar() == 2

    assert Foo().bar() == 1



import logging
import re
import sys
import weakref

import pytest

import aspectlib
import aspectlib.debug

try:
    from StringIO import StringIO
except ImportError:
    from io import StringIO

LOG_TEST_SIMPLE = (
    r'''^some_meth\(1, 2, 3, a=4\) +<<< .*tests/test_aspectlib_debug.py:\d+:test_simple.*
some_meth => \.\.\.\.\.\.\.\.\.\.\.\.\.\.\.\.\.\.\.\.\.\.\.\.\.\.\.\.\.\.\.\. !"#\$%&\'\(\)\*\+,-\./0123456789:;<=>\?@'''
    r'''ABCDEFGHIJKLMNOPQRSTUVWXYZ\[\\\]\^_`abcdefghijklmnopqrstuvwxyz\{\|\}~\.+
$'''
)


def some_meth(*_args, **_kwargs):
    return ''.join(chr(i) for i in range(255))


class MyStuff(object):
    def __init__(self, foo):
        self.foo = foo

    def bar(self):
        return 'foo'

    def stuff(self):
        return self.foo


class OldStuff:
    def __init__(self, foo):
        self.foo = foo

    def bar(self):
        return 'foo'

    def stuff(self):
        return self.foo


def test_simple():
    buf = StringIO()
    with aspectlib.weave(some_meth, aspectlib.debug.log(print_to=buf, module=False, stacktrace=10)):
        some_meth(1, 2, 3, a=4)

    assert re.match(LOG_TEST_SIMPLE, buf.getvalue())
    some_meth(1, 2, 3, a=4)
    assert re.match(LOG_TEST_SIMPLE, buf.getvalue())


def test_fail_to_log():
    @aspectlib.debug.log(print_to="crap")
    def foo():
        pass

    foo()


def test_logging_works():
    buf = StringIO()
    ch = logging.StreamHandler(buf)
    ch.setLevel(logging.DEBUG)
    aspectlib.debug.logger.addHandler(ch)

    @aspectlib.debug.log
    def foo():
        pass

    foo()
    assert re.match(r'foo\(\) +<<<.*\nfoo => None\n', buf.getvalue())


def test_attributes():
    buf = StringIO()
    with aspectlib.weave(
        MyStuff, aspectlib.debug.log(print_to=buf, stacktrace=10, attributes=('foo', 'bar()')), methods='(?!bar)(?!__.*__$)'
    ):
        MyStuff('bar').stuff()
    print(buf.getvalue())
    assert re.match(
        r"^\{test_aspectlib_debug.MyStuff foo='bar' bar='foo'\}.stuff\(\) +<<< .*tests/test_aspectlib_debug.py:\d+:"
        r"test_attributes.*\n\{test_aspectlib_debug.MyStuff foo='bar' bar='foo'\}.stuff => bar\n$",
        buf.getvalue(),
    )
    MyStuff('bar').stuff()
    assert re.match(
        r"^\{test_aspectlib_debug.MyStuff foo='bar' bar='foo'\}.stuff\(\) +<<< .*tests/test_aspectlib_debug.py:\d+:"
        r"test_attributes.*\n\{test_aspectlib_debug.MyStuff foo='bar' bar='foo'\}.stuff => bar\n$",
        buf.getvalue(),
    )


def test_no_stack():
    buf = StringIO()
    with aspectlib.weave(
        MyStuff, aspectlib.debug.log(print_to=buf, stacktrace=None, attributes=('foo', 'bar()')), methods='(?!bar)(?!__.*__$)'
    ):
        MyStuff('bar').stuff()
    print(buf.getvalue())
    assert (
        "{test_aspectlib_debug.MyStuff foo='bar' bar='foo'}.stuff()\n"
        "{test_aspectlib_debug.MyStuff foo='bar' bar='foo'}.stuff => bar\n" == buf.getvalue()
    )


def test_attributes_old_style():
    buf = StringIO()
    with aspectlib.weave(
        OldStuff, aspectlib.debug.log(print_to=buf, stacktrace=10, attributes=('foo', 'bar()')), methods='(?!bar)(?!__.*__$)'
    ):
        OldStuff('bar').stuff()
    print(repr(buf.getvalue()))
    assert re.match(
        r"^\{test_aspectlib_debug.OldStuff foo='bar' bar='foo'\}.stuff\(\) +<<< .*tests/test_aspectlib_debug.py:\d+:"
        r"test_attributes.*\n\{test_aspectlib_debug.OldStuff foo='bar' bar='foo'\}.stuff => bar\n$",
        buf.getvalue(),
    )
    MyStuff('bar').stuff()
    assert re.match(
        r"^\{test_aspectlib_debug.OldStuff foo='bar' bar='foo'\}.stuff\(\) +<<< .*tests/test_aspectlib_debug.py:\d+:"
        r"test_attributes.*\n\{test_aspectlib_debug.OldStuff foo='bar' bar='foo'\}.stuff => bar\n$",
        buf.getvalue(),
    )


def test_no_stack_old_style():
    buf = StringIO()
    with aspectlib.weave(
        OldStuff, aspectlib.debug.log(print_to=buf, stacktrace=None, attributes=('foo', 'bar()')), methods='(?!bar)(?!__.*__$)'
    ):
        OldStuff('bar').stuff()
    print(buf.getvalue())
    assert (
        "{test_aspectlib_debug.OldStuff foo='bar' bar='foo'}.stuff()\n"
        "{test_aspectlib_debug.OldStuff foo='bar' bar='foo'}.stuff => bar\n" == buf.getvalue()
    )


@pytest.mark.skipif(sys.version_info < (2, 7), reason="No weakref.WeakSet on Python<=2.6")
def test_weakref():
    with aspectlib.weave(MyStuff, aspectlib.debug.log):
        s = weakref.WeakSet()
        s.add(MyStuff.stuff)
        print(list(s))
    print(list(s))


@pytest.mark.skipif(sys.version_info < (2, 7), reason="No weakref.WeakSet on Python<=2.6")
def test_weakref_oldstyle():
    with aspectlib.weave(OldStuff, aspectlib.debug.log):
        s = weakref.WeakSet()
        s.add(MyStuff.stuff)
        print(list(s))
    print(list(s))
